import 'dart:async';

import 'package:tuple/tuple.dart';

typedef OnCancel = void Function();

class CanceledError implements Exception {}

//class MergedCancelable2<A, B> implements Future<Tuple2<A, B>> {
//  final Cancelable<A> cancelable1;
//  final Cancelable<B> cancelable2;
//
//  MergedCancelable2(this.cancelable1, this.cancelable2);
//
//  final _result = Completer<Tuple2<A, B>>();
//
//  @override
//  Stream<Tuple2<A, B>> asStream() => Stream.fromFuture(_tupleFuture);
//
//  Future<Tuple2<A, B>> get _tupleFuture =>
//      Future.wait([cancelable1, cancelable2]).then((value) => Tuple2(value.first, value.last));
//
//  @override
//  Future<Tuple2<A, B>> catchError(Function onError, {bool Function(Object error) test}) =>
//      _tupleFuture.catchError(onError, test: test);
//
//  @override
//  Future<R> then<R>(FutureOr<R> Function(Tuple2<A, B> value) onValue, {Function onError}) =>
//      Future.wait([cancelable1, cancelable2])
//          .then((value) => onValue(Tuple2(value.first, value.last)), onError: onError);
//
//  @override
//  Future<Tuple2<A, B>> timeout(Duration timeLimit, {FutureOr<Tuple2<A, B>> Function() onTimeout}) =>
//      _tupleFuture.timeout(timeLimit);
//
//  @override
//  Future<Tuple2<A, B>> whenComplete(FutureOr Function() action) => _tupleFuture.whenComplete(action);
//
//  MergedCancelable2<A, B> next(Function(Tuple2<A, B> value) onValue) {
//    final newMergedCancelable2 = MergedCancelable2(_,_);
//    _tupleFuture.then((value) {
//      try {
//        onValue(value);
//        resultCompleter.complete(value);
//      } catch (error) {
//        resultCompleter.completeError(error);
//      }
//    }, onError: (e) {
//      resultCompleter.completeError(e);
//    });
//    return MergedCancelable2(resultCompleter, cancel);
//  }
//}

class Cancelable<O> implements Future<O> {
  final Completer<O> _completer;
  final OnCancel onCancel;

  Cancelable(this._completer, this.onCancel);

  Future<O> get _future => _completer.future;

  void cancel() {
    onCancel();
    if (!_completer.isCompleted) _completer.completeError(CanceledError());
  }

  Cancelable<O> next(Function(O value) onValue) {
    final resultCompleter = Completer<O>();
    _future.then((value) {
      try {
        onValue(value);
        resultCompleter.complete(value);
      } catch (error) {
        resultCompleter.completeError(error);
      }
    }, onError: (e) {
      resultCompleter.completeError(e);
    });
    return Cancelable(resultCompleter, cancel);
  }

  @override
  Stream<O> asStream() => _future.asStream();

  @override
  Future<O> catchError(Function onError, {bool Function(Object error) test}) => _future.catchError(onError, test: test);

  @override
  Future<R> then<R>(Function(O value) onValue, {Function onError}) => _future.then(onValue, onError: onError);

  @override
  Future<O> timeout(Duration timeLimit, {FutureOr Function() onTimeout}) => _future.timeout(timeLimit);

  @override
  Future<O> whenComplete(FutureOr Function() action) => _future.whenComplete(action);
}
